case class Identifier(name: String) {
  override def toString = name
}

sealed abstract class Expr {
  def subst(m: Map[Identifier, Expr]): Expr = this match {
    case BooleanLiteral(_) => this
    case BigIntLiteral(_) => this
    case Var(id) if m.contains(id) => m(id)
    case Var(_) => this
    case Old(_) => this
    case UnaryMinus(e) => UnaryMinus(e.subst(m))
    case Plus(e1, e2) => Plus(e1.subst(m), e2.subst(m))
    case Minus(e1, e2) => Minus(e1.subst(m), e2.subst(m))
    case Mult(e1, e2) => Mult(e1.subst(m), e2.subst(m))
    case Mod(e1, e2) => Mod(e1.subst(m), e2.subst(m))
    case Div(e1, e2) => Div(e1.subst(m), e2.subst(m))
    case Lt(e1, e2) => Lt(e1.subst(m), e2.subst(m))
    case Leq(e1, e2) => Leq(e1.subst(m), e2.subst(m))
    case Gt(e1, e2) => Gt(e1.subst(m), e2.subst(m))
    case Geq(e1, e2) => Geq(e1.subst(m), e2.subst(m))
    case Equals(e1, e2) => Equals(e1.subst(m), e2.subst(m))
    case And(e1, e2) => And(e1.subst(m), e2.subst(m))
    case Or(e1, e2) => Or(e1.subst(m), e2.subst(m))
    case Implies(e1, e2) => Implies(e1.subst(m), e2.subst(m))
    case Not(e) => Not(e.subst(m))
  }

  def replace(m: Map[Expr, Expr]): Expr = {
    if (m.contains(this)) m(this)
    else {
      this match {
        case BooleanLiteral(_) => this
        case BigIntLiteral(_) => this
        case Var(_) => this
        case Old(_) => this
        case UnaryMinus(e) => UnaryMinus(e.replace(m))
        case Plus(e1, e2) => Plus(e1.replace(m), e2.replace(m))
        case Minus(e1, e2) => Minus(e1.replace(m), e2.replace(m))
        case Mult(e1, e2) => Mult(e1.replace(m), e2.replace(m))
        case Mod(e1, e2) => Mod(e1.replace(m), e2.replace(m))
        case Div(e1, e2) => Div(e1.replace(m), e2.replace(m))
        case Lt(e1, e2) => Lt(e1.replace(m), e2.replace(m))
        case Leq(e1, e2) => Leq(e1.replace(m), e2.replace(m))
        case Gt(e1, e2) => Gt(e1.replace(m), e2.replace(m))
        case Geq(e1, e2) => Geq(e1.replace(m), e2.replace(m))
        case Equals(e1, e2) => Equals(e1.replace(m), e2.replace(m))
        case And(e1, e2) => And(e1.replace(m), e2.replace(m))
        case Or(e1, e2) => Or(e1.replace(m), e2.replace(m))
        case Implies(e1, e2) => Implies(e1.replace(m), e2.replace(m))
        case Not(e) => Not(e.replace(m))
      }
    }
  }
}

case class BooleanLiteral(b: Boolean) extends Expr {
  override def toString = b.toString
}

case class BigIntLiteral(i: BigInt) extends Expr {
  override def toString = i.toString
}

case class Var(id: Identifier) extends Expr {
  override def toString = id.toString
}

case class Old(id: Identifier) extends Expr {
  override def toString = s"old($id)"
}

case class UnaryMinus(e: Expr) extends Expr {
  override def toString = s"-$e"
}

case class Plus(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 + $e2)"
}

case class Minus(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 - $e2)"
}

case class Mult(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 * $e2)"
}

case class Mod(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 % $e2)"
}

case class Div(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 / $e2)"
}

case class Lt(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 < $e2)"
}

case class Leq(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 <= $e2)"
}

case class Gt(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 > $e2)"
}

case class Geq(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 >= $e2)"
}

case class Equals(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 == $e2)"
}

case class Not(e: Expr) extends Expr {
  override def toString = s"!$e"
}

case class And(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 && $e2)"
}

case class Or(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 || $e2)"
}

case class Implies(e1: Expr, e2: Expr) extends Expr {
  override def toString = s"($e1 ==> $e2)"
}

object Expr {
  def and(es: Iterable[Expr]): Expr = {
    if (es.isEmpty) BooleanLiteral(true)
    else es.tail.foldLeft(es.head)(And(_,_))
  }

  def and(e: Expr, es: Expr*): Expr = {
    es.foldLeft(e)(And(_,_))
  }

  def or(es: Iterable[Expr]): Expr = {
    if (es.isEmpty) BooleanLiteral(false)
    else es.tail.foldLeft(es.head)(Or(_,_))
  }

  def or(e: Expr, es: Expr*): Expr = {
    es.foldLeft(e)(Or(_,_))
  }
}
