import scallion.syntactic._

import java.io.File

object ProgramParser extends Syntaxes[Token, Kind] with Operators {

  import SafeImplicits._

  override def getKind(token: Token): Kind = Kind.of(token)

  val comma: Syntax[Unit] = elem(Kind.Comma).unit()
  val open: Syntax[Unit] = elem(Kind.Parenthesis(true)).unit()
  val close: Syntax[Unit] = elem(Kind.Parenthesis(false)).unit()
  val equal: Syntax[Unit] = elem(Kind.Equal).unit()

  val lpar = elem(Kind.Parenthesis(true)).unit()
  val rpar = elem(Kind.Parenthesis(false)).unit()
  val lbra = elem(Kind.CurlyBracket(true)).unit()
  val rbra = elem(Kind.CurlyBracket(false)).unit()
  val lsqbra = elem(Kind.SquareBracket(true)).unit()
  val rsqbra = elem(Kind.SquareBracket(false)).unit()

  val plus = elem(Kind.Plus)
  val minus = elem(Kind.Minus)
  val mult = elem(Kind.Mult)
  val mod = elem(Kind.Mod)
  val div = elem(Kind.Div)

  val not = elem(Kind.Not).unit()
  val and = elem(Kind.And)
  val or = elem(Kind.Or)
  val implies = elem(Kind.Implies)

  val lt = elem(Kind.Lt)
  val leq = elem(Kind.Leq)
  val gt = elem(Kind.Gt)
  val geq = elem(Kind.Geq)
  val doubleEqual = elem(Kind.DoubleEqual)
  val notEqual = elem(Kind.NotEqual)

  val defK = elem(Kind.Def).unit()
  val ensuringK = elem(Kind.Ensuring).unit()
  val varK = elem(Kind.Var).unit()
  val ifK = elem(Kind.If).unit()
  val elseK = elem(Kind.Else).unit()
  val invariantK = elem(Kind.Invariant).unit()
  val oldK = elem(Kind.Old).unit()
  val havocK = elem(Kind.Havoc).unit()
  val assumeK = elem(Kind.Assume).unit()

  val colon = elem(Kind.Colon).unit()

  val id: Syntax[Identifier] = elem(Kind.Identifier).map {
    case Token.Identifier(x, _) => Identifier(x)
    case _ => throw new Exception("Unreachable code (idExpr)")
  }

  val tpe: Syntax[Type] =
    elem(Kind.BoolType).map(_ => BoolType: Type) |
    elem(Kind.BigIntType).map(_ => BigIntType: Type)

  val boolLiteral: Syntax[Expr] = elem(Kind.BooleanLiteral).map {
    case Token.BooleanLiteral(b, _) => BooleanLiteral(b)
    case _ => throw new Exception("Unreachable code (booleanLiteral)")
  }

  val intLiteral: Syntax[Expr] = elem(Kind.BigIntLiteral).map {
    case Token.BigIntLiteral(i, _) => BigIntLiteral(i)
    case _ => throw new Exception("Unreachable code (intLiteral)")
  }

  val idExpr: Syntax[Expr] = id.map(Var(_))

  val oldExpr: Syntax[Expr] = (oldK ~>~ lpar ~>~ id ~<~ rpar).map(Old(_))

  lazy val basicExpr: Syntax[Expr] = boolLiteral | intLiteral | idExpr | oldExpr
  lazy val simpleExpr: Syntax[Expr] = basicExpr | lpar ~>~ expr ~<~ rpar

  lazy val minusExpr: Syntax[Expr] = prefixes(minus, simpleExpr) {
    case (_, e) => UnaryMinus(e)
  }

  lazy val minusNotExpr: Syntax[Expr] = prefixes(not, simpleExpr) {
    case (_, e) => Not(e)
  }

  lazy val expr: Syntax[Expr] = recursive {
    operators(minusExpr)(
      mod is LeftAssociative,
      mult | div is LeftAssociative,
      plus | minus is LeftAssociative,
      lt | gt | leq | geq is LeftAssociative,
      doubleEqual | notEqual is LeftAssociative,
      and is LeftAssociative,
      or is LeftAssociative,
      implies is RightAssociative,
    )
    {
      case (x, Token.Implies(_), y) => Implies(x, y)
      case (x, Token.And(_), y) => And(x, y)
      case (x, Token.Or(_), y) => Or(x, y)
      case (x, Token.Plus(_), y) => Plus(x, y)
      case (x, Token.Minus(_), y) => Minus(x, y)
      case (x, Token.Mult(_), y) => Mult(x, y)
      case (x, Token.Div(_), y) => Div(x, y)
      case (x, Token.Mod(_), y) => Mod(x, y)
      case (x, Token.Lt(_), y) => Lt(x, y)
      case (x, Token.Gt(_), y) => Gt(x, y)
      case (x, Token.Leq(_), y) => Leq(x, y)
      case (x, Token.Geq(_), y) => Geq(x, y)
      case (x, Token.NotEqual(_), y) => Not(Equals(x, y))
      case (x, Token.DoubleEqual(_), y) => Equals(x, y)
      case _ => throw new Exception("Unreachable code (expr)")
    }
  }

  val havoc: Syntax[Statement] = (havocK ~>~ id).map(Havoc(_))
  val assume: Syntax[Statement] = (assumeK ~>~ expr).map(Assume(_))

  val ite: Syntax[Statement] = {
    (ifK ~
      lpar ~ expr ~ rpar ~
      lbra ~ statement ~ rbra ~
      opt(elseK ~>~ lbra ~>~ statement ~<~ rbra)
    ).map {
      case
        _ ~ _ ~ e ~ _ ~
        _ ~ c1 ~ _ ~
        c2 => IfThenElse(e, c1, c2.getOrElse(NoOp))
    }
  }

  val invArgs: Syntax[Seq[Expr]] = lpar ~>~ repsep(expr, comma) ~<~ rpar

  val assignment: Syntax[Expr] = equal ~>~ expr

  val funInvOrAssignment: Syntax[Statement] = {
    (id ~ (invArgs || assignment)).map {
      case f ~ Left(args) => ProcedureInvocation(f, args)
      case x ~ Right(e) => Assignment(x, e)
    }
  }

  val basicStatement: Syntax[Statement] =
    ite | funInvOrAssignment | havoc | assume

  lazy val statement: Syntax[Statement] = recursive {
    many(basicStatement).map { xs =>
      if (xs.isEmpty) NoOp
      else xs.tail.foldLeft(xs.head)(Compose(_, _))
    }
  }

  val argument: Syntax[Arg] = {
    id ~ colon ~ tpe
  }.map {
    case x ~ _ ~ t => Arg(x, t)
  }

  val ensuring: Syntax[Expr] = ensuringK ~>~ lpar ~>~ expr ~<~ rpar

  val procDef: Syntax[ProcDef] = {
    (defK ~ id ~
      lpar ~ repsep(argument, comma.unit()) ~ rpar ~
      lbra ~ statement ~ rbra ~
      opt(ensuring)
    ).map {
      case
        _ ~ f ~
        _ ~ args ~ _ ~
        _ ~ body ~ _ ~
        post =>
        ProcDef(f, args, body, post)
    }
  }

  val varDef: Syntax[VarDef] = {
    (varK ~ id ~ colon ~ tpe).map {
      case _ ~ x ~ _ ~ t =>
        VarDef(x, t)
    }
  }

  val program: Syntax[Program] =
    (many(varDef) ~ many(procDef)).map {
      case vds ~ procs =>
        Program(
          vds,
          procs.map((proc: ProcDef) => (proc.id, proc)).toMap
        )
    }

  assert(program.isLL1, "Program syntax should be LL1")
  assert(program.isProductive, "Program syntax should be productive")

  def apply(it: Iterator[Token]): ParseResult[Program] = program(it)
}

object Parser {
  def parseString(s: String): Program = {
    ProgramParser(Lexer(s.iterator)) match {
      case ProgramParser.Parsed(value, _) =>
        value
      case ProgramParser.UnexpectedEnd(_) =>
        throw new Exception("Error during parsing: unexpected end.")
      case ProgramParser.UnexpectedToken(t, _) =>
        throw new Exception(s"Error during parsing: unexpected token: $t")
    }
  }

  def parseFile(f: File): Program = {
    parseString(scala.io.Source.fromFile(f).getLines.mkString("\n"))
  }
}
