import Utils._

sealed abstract class Statement

case object NoOp extends Statement {
  override def toString = ""
}

case class ProcedureInvocation(id: Identifier, args: Seq[Expr]) extends Statement {
  override def toString = s"$id(${args.mkString(", ")})"
}

case class IfThenElse(e: Expr, c1: Statement, c2: Statement) extends Statement {
  override def toString = {
    s"if ($e) {\n${indent(c1.toString)}\n}" +
    (if (c2 == NoOp) "" else s"else {\n${indent(c2.toString)}\n}")
  }
}

case class Assignment(x: Identifier, e: Expr) extends Statement {
  override def toString = s"$x = $e"
}

case class Compose(c1: Statement, c2: Statement) extends Statement {
  override def toString = Seq(c1, c2).filterNot(_ == NoOp).mkString("\n")
}

case class Havoc(id: Identifier) extends Statement {
  override def toString = s"havoc $id"
}

case class Assume(e: Expr) extends Statement {
  override def toString = s"assume $e"
}

sealed abstract class Type

case object BoolType extends Type {
  override def toString = "Boolean"
}

case object BigIntType extends Type {
  override def toString = "Int"
}

case class Arg(id: Identifier, tpe: Type) {
  override def toString = {
    s"$id: $tpe"
  }
}

case class ProcDef(
  id: Identifier,
  args: Seq[Arg],
  body: Statement,
  ensuring: Option[Expr]
) {
  override def toString = {
    s"def $id(${args.mkString(", ")}) {\n${indent(body.toString)}\n}" +
    ensuring.map(e => s" ensuring($e)").mkString
  }
}

case class VarDef(id: Identifier, tpe: Type) {
  override def toString = {
    s"var $id: $tpe"
  }
}

case class Program(
  val variables: Seq[VarDef],
  val procedures: Map[Identifier, ProcDef]
) {
  require(procedures.values.forall(proc => proc.args.map(_.id).intersect(variables.map(_.id)).isEmpty),
    "Please name procedure arguments differently than global variables."
  )
  require(variables.toSet.size == variables.size,
    "Program has two global variables with the same name"
  )
  require(procedures.values.forall(proc => proc.args.toSet.size == proc.args.size),
    "A procedure has two arguments with the same name"
  )

  override def toString =
    variables.mkString("\n") + "\n\n" +
    procedures.values.mkString("\n\n")
}
